/*
 * Licensed Materials - Property of IBM
 * 5724-Y62 WebSphere Sensor Events
 * (c) Copyright IBM Corp. 2009  All rights reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM
 * Corp.
 *
 * DISCLAIMER OF WARRANTIES.  The following code is sample code
 * created by IBM Corporation.  This sample code is part of
 * WebSphere Sensor Events and is warranted to perform its
 * intended function only if used unmodified.  If you modify
 * this code then it is considered provided "AS IS", without
 * warranty of any kind.  Notwithstanding the foregoing, IBM
 * shall not be liable for any damages arising out of your use
 * of the sample code, even if they have been advised of the
 * possibility of such damages.
 */

package com.ibm.sensorevent.persistence.sample.ejb;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TimeZone;
import java.util.TreeMap;

import javax.naming.InitialContext;
import javax.sql.DataSource;

import com.ibm.rfid.premises.app.access.SensorEventsAPIAccess;
import com.ibm.rfid.premises.app.access.SensorEventsAPIContext;
import com.ibm.rfid.premises.entities.Location;
import com.ibm.sensorevent.model.Group;
import com.ibm.sensorevent.model.IHeader;
import com.ibm.sensorevent.model.IPayload;
import com.ibm.sensorevent.model.IPayloadMetaData;
import com.ibm.sensorevent.model.ISensorEvent;
import com.ibm.sensorevent.model.payload.EPCTag;
import com.ibm.sensorevent.model.payload.ITag;
import com.ibm.sensorevent.model.payload.PassiveRFIDAggregatedTagReadPayload;
import com.ibm.sensorevent.model.payload.PassiveRFIDTagRead;
import com.ibm.sensorevent.model.payload.PassiveRFIDTagReadPayload;
import com.ibm.sensorevent.model.payload.Tag;
import com.ibm.sensorevent.model.payload.TagReadFeedbackPayload;

/**
 * Bean implementation class for Enterprise Bean: SamplePersistence
 */
public class SamplePersistenceBean extends com.ibm.sensorevent.engine.baseagent.IBMSEAbstractTaskAgent
implements javax.ejb.MessageDrivenBean, javax.jms.MessageListener {
	public static final String COPYRIGHT = 
		"Licensed Materials - Property of IBM " +
		"5724-Y62 WebSphere Sensor Events " +
		"(c) Copyright IBM Corp. 2009  All rights reserved. " +
		"US Government Users Restricted Rights - Use, duplication or " +
		"disclosure restricted by GSA ADP Schedule Contract with IBM " +
		"Corp.";
	private static final long serialVersionUID = 1L;
	private javax.ejb.MessageDrivenContext fMessageDrivenCtx;
	
	public static final String JDBC_REF = "java:comp/env/jdbc/dbref";
	
	public static final String ATTR_TAGREADFEEDBACK = "tagreadfeedback";
	
	public static final String SQL_SELECT_HEADER =
		"select event_id from sample.persistence_header where event_id=?";
			
	public static final String SQL_INSERT_HEADER =
		"insert into sample.persistence_header (event_id, source_id, source_name, event_type, event_date_time) values (?, ?, ?, ?, ?)";
	
	public static final String SQL_INSERT_METADATA =
		"insert into sample.persistence_metadata (event_id, attr_name, attr_value) values (?, ?, ?)";
	
	public static final String SQL_UPDATE_METADATA =
		"update sample.persistence_metadata set attr_value=? where event_id=? and attr_name=?";
	
	public static final String SQL_INSERT_PAYLOAD =
		"insert into sample.persistence_payload (event_id, tag_iduri, tag_rawuri, tag_tagid, tag_taguri) values (?, ?, ?, ?, ?)";
	
	public static final String SQL_INSERT_PAYLOAD_EXT =
		"insert into sample.persistence_payload_ext (event_id, tag_tagid, attr_name, attr_value) values (?, ?, ?, ?)";
	
	public static final String SQL_UPDATE_PAYLOAD_EXT =
		"update sample.persistence_payload_ext set attr_value=? where event_id=? and tag_tagid=? and attr_name=?";

	private DataSource dataSource;
	
	/* (non-Javadoc)
	 * @see com.ibm.sensorevent.engine.baseagent.IBMSEAbstractTaskAgent#onIBMSensorEvent(com.ibm.sensorevent.model.ISensorEvent)
	 */
	@Override
	protected void onIBMSensorEvent(ISensorEvent ise) {
		Connection con = null;
		
		try {
			IPayload payload = ise.getPayload();
			if (payload instanceof PassiveRFIDTagReadPayload || payload instanceof PassiveRFIDAggregatedTagReadPayload) {
				con = this.dataSource.getConnection();
				con.setAutoCommit(false);
				
				String eventId = ise.getHeader().getEventId();
				
				// process the header
				this.handleHeader(con, ise.getHeader());
				
				// process the meta data
				this.handleMetaData(con, eventId, ise.getPayloadMetaData());
				
				// process the payload
				this.handlePayload(con, eventId, payload);
				
				con.commit();
			} else if (payload instanceof TagReadFeedbackPayload) {
				con = this.dataSource.getConnection();
				con.setAutoCommit(false);
				
				// use the originating event ID
				String origEventId = ise.getHeader().getOriginatingEventId();
				
				// has the originating event ID already been persisted?
				if (this.doesEventExist(con, origEventId)) {
					// only process the payload if the event already exists
					this.handleFeedbackPayload(con, origEventId, (TagReadFeedbackPayload) payload);
				}
				
				con.commit();
			}
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			try {
				con.close();
			} catch (Exception e) {
				// no-op
			}
		}
	}
	
	private boolean doesEventExist(Connection con, String eventId) throws Exception {
		PreparedStatement ps = null;
		ResultSet rs = null;
		try {
			ps = con.prepareStatement(SQL_SELECT_HEADER);
			
			ps.setString(1, eventId);
			
			rs = ps.executeQuery();
			boolean result = false;
			while (rs.next()) {
				result = true;
				break;
			}
			return result;
		} catch (Exception e) {
			throw e;
		} finally {
			try {
				rs.close();
			} catch (Exception e) {
				// no-op
			}
			
			try {
				ps.close();
			} catch (Exception e) {
				// no-op
			}
		}	
	}
	
	private void handleHeader(Connection con, IHeader header) throws Exception {
		String eventId = header.getEventId();
		String sourceId = header.getSourceId();
		String eventType = header.getEventType();
		Timestamp eventDateTime = new Timestamp(header.getDateTime());
		
		String sourceName = sourceId;
		try {
			// convert sourceId to sourceName
			SensorEventsAPIAccess wseAccess = SensorEventsAPIContext.getSensorEventsAPIAccess();
			Location location = wseAccess.getLocation(sourceId);
			sourceName = location.getUserName();
		} catch (Exception e) {
			// default to sourceId if an exception occurs
			sourceName = sourceId;
		}
		
		PreparedStatement ps = null;
		
		try {			
			ps = con.prepareStatement(SQL_INSERT_HEADER);
			
			ps.setString(1, eventId);
			ps.setString(2, sourceId);
			ps.setString(3, sourceName);
			ps.setString(4, eventType);
			
			Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
			ps.setTimestamp(5, eventDateTime, cal);
			
			ps.executeUpdate();
		} catch (Exception e) {
			throw e;
		} finally {
			try {
				ps.close();
			} catch (Exception e) {
				// no-op
			}
		}		
	}
	
	@SuppressWarnings("unchecked")
	private void handleMetaData(Connection con, String eventId, IPayloadMetaData pmd) throws Exception {
		Map<String, String> pmdAttrs = new TreeMap<String, String>(pmd.toMap());
		
		if (pmdAttrs != null && !pmdAttrs.isEmpty()) {
			for (Iterator<Map.Entry<String, String>> i=pmdAttrs.entrySet().iterator(); i.hasNext(); ) {
				Map.Entry<String, String> me = i.next();
				String attrName = me.getKey();
				String attrValue = me.getValue();
				
				try {
					this.insertMetaData(con, eventId, attrName, attrValue);
				} catch (Exception e) {
					// insert failed; try updating
					try {
						this.updateMetaData(con, eventId, attrName, attrValue);
					} catch (Exception e2) {
						// just print the stack trace and continue to next attribute
						e2.printStackTrace();
					}
				}
			}
		}
	}
	
	private void insertMetaData(Connection con, String eventId, String attrName, String attrValue) throws Exception {
		PreparedStatement ps = null;
		try {
			ps = con.prepareStatement(SQL_INSERT_METADATA);
				
			ps.setString(1, eventId);
			ps.setString(2, attrName);
			ps.setString(3, attrValue);
				
			ps.executeUpdate();
		} catch (Exception e) {
			throw e;
		} finally {
			try {
				ps.close();
			} catch (Exception e) {
				// no-op
			}
		}
	}
	
	private void updateMetaData(Connection con, String eventId, String attrName, String attrValue) throws Exception {
		PreparedStatement ps = null;
		try {
			ps = con.prepareStatement(SQL_UPDATE_METADATA);
				
			ps.setString(1, attrValue);
			ps.setString(2, eventId);
			ps.setString(3, attrName);
				
			ps.executeUpdate();
		} catch (Exception e) {
			throw e;
		} finally {
			try {
				ps.close();
			} catch (Exception e) {
				// no-op
			}
		}
	}
	
	private void handlePayload(Connection con, String eventId, IPayload payload) throws Exception {
		List<PassiveRFIDTagRead> tagReads = this.getTagReads(payload);
		
		if (tagReads != null && !tagReads.isEmpty()) {
			PreparedStatement ps = null;
			try {				
				ps = con.prepareStatement(SQL_INSERT_PAYLOAD);
				
				for (Iterator<PassiveRFIDTagRead> i=tagReads.iterator(); i.hasNext(); ) {
					PassiveRFIDTagRead tagRead = i.next();
					
					EPCTag epcTag = this.extractEpcTag(tagRead);
					if (epcTag != null) {
						String tagId = epcTag.getTagID();
						
						ps.setString(1, eventId);
						ps.setString(2, epcTag.getAsIDURI());
						ps.setString(3, epcTag.getAsRawURI());
						ps.setString(4, tagId);
						ps.setString(5, epcTag.getAsTagURI());
						
						try {
							ps.executeUpdate();
							
							this.handlePayloadExt(con, eventId, tagId, tagRead);
						} catch (Exception e) {
							 e.printStackTrace();
						}
					}
				}
			} catch (Exception e) {
				throw e;
			} finally {
				try {
					ps.close();
				} catch (Exception e) {
					// no-op
				}
			}
		}
	}
	
	private void handleFeedbackPayload(Connection con, String eventId, TagReadFeedbackPayload payload) throws Exception {
		String attrName = ATTR_TAGREADFEEDBACK;
		String attrValue = payload.getValue();
		
		try {
			this.insertMetaData(con, eventId, attrName, attrValue);
		} catch (Exception e) {
			// insert failed; try updating
			this.updateMetaData(con, eventId, attrName, attrValue);
		}
	}
	
	@SuppressWarnings("unchecked")
	private void handlePayloadExt(Connection con, String eventId, String tagId, PassiveRFIDTagRead tagRead) throws Exception {
		Map<String, String> tagUserData = ((Group) ((Tag) tagRead.getTag()).getGroup(ITag.TAGUSERDATA)).toMap();
		
		if (tagUserData != null && !tagUserData.isEmpty()) {
			for (Iterator<Map.Entry<String, String>> i=tagUserData.entrySet().iterator(); i.hasNext(); ) {
				Map.Entry<String, String> me = i.next();
				String attrName = me.getKey();
				String attrValue = me.getValue();
			
				try {
					this.insertPayloadExt(con, eventId, tagId, attrName, attrValue);
				} catch (Exception e) {
					// insert failed; try updating
					try {
						this.updatePayloadExt(con, eventId, tagId, attrName, attrValue);
					} catch (Exception e2) {
						// just print the stack trace and continue to next attribute
						e2.printStackTrace();
					}
				}
			}	
		}
	}
	
	private void insertPayloadExt(Connection con, String eventId, String tagId, String attrName, String attrValue) throws Exception {
		PreparedStatement ps = null;
		try {
			ps = con.prepareStatement(SQL_INSERT_PAYLOAD_EXT);

			ps.setString(1, eventId);
			ps.setString(2, tagId);
			ps.setString(3, attrName);
			ps.setString(4, attrValue);
				
			ps.executeUpdate();
		} catch (Exception e) {
			throw e;
		} finally {
			try {
				ps.close();
			} catch (Exception e) {
				// no-op
			}
		}
	}
	
	private void updatePayloadExt(Connection con, String eventId, String tagId, String attrName, String attrValue) throws Exception {
		PreparedStatement ps = null;
		try {
			ps = con.prepareStatement(SQL_UPDATE_PAYLOAD_EXT);

			ps.setString(1, attrValue);
			ps.setString(2, eventId);
			ps.setString(3, tagId);
			ps.setString(4, attrName);
				
			ps.executeUpdate();
		} catch (Exception e) {
			throw e;
		} finally {
			try {
				ps.close();
			} catch (Exception e) {
				// no-op
			}
		}
	}
	
	@SuppressWarnings("unchecked")
	private List<PassiveRFIDTagRead> getTagReads(IPayload payload) throws Exception {
		List<PassiveRFIDTagRead> tagReads = new ArrayList<PassiveRFIDTagRead>();
		
		if (payload instanceof PassiveRFIDTagReadPayload) {
			PassiveRFIDTagReadPayload tagReadPayload = (PassiveRFIDTagReadPayload) payload;
			PassiveRFIDTagRead prtr = (PassiveRFIDTagRead) tagReadPayload.getTagRead();
			tagReads.add(prtr);
		} else if (payload instanceof PassiveRFIDAggregatedTagReadPayload) {
			PassiveRFIDAggregatedTagReadPayload aggTagReadPayload = (PassiveRFIDAggregatedTagReadPayload) payload;
			tagReads.addAll(aggTagReadPayload.getTagReads());			
		}
		
		return tagReads;
	}
	
	private EPCTag extractEpcTag(PassiveRFIDTagRead tagRead) throws Exception {
		if (tagRead.getTag() instanceof EPCTag) {
			EPCTag epcTag = (EPCTag) tagRead.getTag();
			return epcTag;
		}
		
		return null;
	}
	
	/**
	 * getMessageDrivenContext
	 */
	public javax.ejb.MessageDrivenContext getMessageDrivenContext() {
		return fMessageDrivenCtx;
	}

	/**
	 * setMessageDrivenContext
	 */
	public void setMessageDrivenContext(javax.ejb.MessageDrivenContext ctx) {
		fMessageDrivenCtx = ctx;
	}

	/**
	 * ejbCreate
	 */
	public void ejbCreate() {
		try {
			InitialContext ic = new InitialContext();
			this.dataSource = (DataSource) ic.lookup(JDBC_REF);
		} catch (Exception e) {
			e.printStackTrace();
			this.dataSource = null;
		}
	}

	/**
	 * onMessage
	 */
	public void onMessage(javax.jms.Message msg) {
		super.onMessage(msg);
	}

	/**
	 * ejbRemove
	 */
	public void ejbRemove() {
	}

}
